/**
  *  \file planetscentral/dbclient/dbset.cc
  *  \brief Database Client: Sets
  */

#include "planetscentral/dbclient/dbset.h"
#include "planetscentral/dbclient/dbclient.h"

/** Constructor.
    \param client Database client
    \param name Name of key */
DbSet::DbSet(DbClient& client, const string_t& name)
    : DbKey(client, name)
{ }

/** Get number of elements in set. */
int32_t
DbSet::size()
{
    return client.callInt(DbRequest().withString("SCARD").withString(key));
}

/** Check for empty set. */
bool
DbSet::empty()
{
    return size() == 0;
}

/** Sort operation on this set. */
DbSort
DbSet::sort()
{
    return DbSort(getClient(), getName());
}

/******************************** DbIntSet *******************************/

/** Constructor.
    \param client Database client
    \param name Name of key */
DbIntSet::DbIntSet(DbClient& client, const string_t& name)
    : DbSet(client, name)
{ }

/** Add a value to the set.
    \param value Value to add
    \retval true element was added
    \retval false element was already in the set */
bool
DbIntSet::add(int32_t value)
{
    return client.callInt(DbRequest().withString("SADD").withString(key).withInt(value));
}

/** Remove a value from the set.
    \param value Value to remove
    \retval true element was removed
    \retval false element was not in the set */
bool
DbIntSet::remove(int32_t value)
{
    return client.callInt(DbRequest().withString("SREM").withString(key).withInt(value));
}

/** Extract a random value. Get a random element and remove it from the set. */
int32_t
DbIntSet::extractRandom()
{
    return client.callInt(DbRequest().withString("SPOP").withString(key));
}

/** Get a random value. */
int32_t
DbIntSet::getRandom()
{
    return client.callInt(DbRequest().withString("SRANDMEMBER").withString(key));
}

/** Move a value into another set. It is not an error if the receiving set
    already contains the element.
    \param value Value to move
    \param other Receiving set
    \retval true element was moved
    \retval false element was not contained in this set */
bool
DbIntSet::moveTo(int32_t value, const DbIntSet& other)
{
    return client.callInt(DbRequest().withString("SMOVE").withString(key).withString(other.key).withInt(value));
}

/** Check whether set contains an element.
    \param value Element to check */
bool
DbIntSet::contains(int32_t value)
{
    return client.callInt(DbRequest().withString("SISMEMBER").withString(key).withInt(value));
}

/** Get all elements. Order will be unspecified; use sort() to define an order.
    \param list [out] Set elements will be appended here */
void
DbIntSet::getAll(IntList_t& list)
{
    std::auto_ptr<IntValue> val(client.call(DbRequest().withString("SMEMBERS").withString(key)));
    toIntListResult(val.get(), list);
}

/** Compute intersection of sets.
    Use .andAlso() to intersect more sets, .get() to get the result, .storeTo() to store it. */
DbIntSetOperation
DbIntSet::intersect(DbIntSet other)
{
    return DbIntSetOperation("SINTER", *this, other);
}

/** Compute union of sets.
    Use .andAlso() to merge more sets, .get() to get the result, .storeTo() to store it. */
DbIntSetOperation
DbIntSet::merge(DbIntSet other)
{
    return DbIntSetOperation("SUNION", *this, other);
}

/** Compute difference of sets.
    Use .andAlso() to subtract more sets, .get() to get the result, .storeTo() to store it. */
DbIntSetOperation
DbIntSet::remove(DbIntSet other)
{
    return DbIntSetOperation("SDIFF", *this, other);
}

/****************************** DbStringSet ******************************/

/** Constructor.
    \param client Database client
    \param name Name of key */
DbStringSet::DbStringSet(DbClient& client, const string_t& name)
    : DbSet(client, name)
{ }

/** Add a value to the set.
    \param value Value to add
    \retval true element was added
    \retval false element was already in the set */
bool
DbStringSet::add(string_t value)
{
    return client.callInt(DbRequest().withString("SADD").withString(key).withString(value));
}

/** Remove a value from the set.
    \param value Value to remove
    \retval true element was removed
    \retval false element was not in the set */
bool
DbStringSet::remove(string_t value)
{
    return client.callInt(DbRequest().withString("SREM").withString(key).withString(value));
}

/** Extract a random value. Get a random element and remove it from the set. */
string_t
DbStringSet::extractRandom()
{
    return client.callString(DbRequest().withString("SPOP").withString(key));
}

/** Get a random value. */
string_t
DbStringSet::getRandom()
{
    return client.callString(DbRequest().withString("SRANDMEMBER").withString(key));
}

/** Move a value into another set. It is not an error if the receiving set
    already contains the element.
    \param value Value to move
    \param other Receiving set
    \retval true element was moved
    \retval false element was not contained in this set */
bool
DbStringSet::moveTo(string_t value, const DbStringSet& other)
{
    return client.callInt(DbRequest().withString("SMOVE").withString(key).withString(other.key).withString(value));
}

/** Check whether set contains an element.
    \param value Element to check */
bool
DbStringSet::contains(string_t value)
{
    return client.callInt(DbRequest().withString("SISMEMBER").withString(key).withString(value));
}

/** Get all elements. Order will be unspecified; use sort() to define an order.
    \param list [out] Set elements will be appended here */
void
DbStringSet::getAll(StringList_t& list)
{
    std::auto_ptr<IntValue> val(client.call(DbRequest().withString("SMEMBERS").withString(key)));
    toStringListResult(val.get(), list);
}

/** Compute intersection of sets.
    Use .andAlso() to intersect more sets, .get() to get the result, .storeTo() to store it. */
DbStringSetOperation
DbStringSet::intersect(DbStringSet other)
{
    return DbStringSetOperation("SINTER", *this, other);
}

/** Compute union of sets.
    Use .andAlso() to merge more sets, .get() to get the result, .storeTo() to store it. */
DbStringSetOperation
DbStringSet::merge(DbStringSet other)
{
    return DbStringSetOperation("SUNION", *this, other);
}

/** Compute difference of sets.
    Use .andAlso() to subtract more sets, .get() to get the result, .storeTo() to store it. */
DbStringSetOperation
DbStringSet::remove(DbStringSet other)
{
    return DbStringSetOperation("SDIFF", *this, other);
}

/*************************** DbIntSetOperation ***************************/

/** Constructor.
    \param op Operation
    \param a,b Arguments */
DbIntSetOperation::DbIntSetOperation(const char* op, DbIntSet a, DbIntSet b)
    : client(a.getClient()),
      req(new DbRequest())
{
    req->withString(op).withString(a.getName()).withString(b.getName());
}

/** Add an additional argument. */
DbIntSetOperation&
DbIntSetOperation::andAlso(DbIntSet x)
{
    req->withString(x.getName());
    return *this;
}

/** Get the result. */
void
DbIntSetOperation::get(IntList_t& list)
{
    std::auto_ptr<IntValue> result(client.call(*req));
    toIntListResult(result.get(), list);
}

/** Store the result in another set. */
void
DbIntSetOperation::storeTo(DbIntSet other)
{
    req->setNew(0, makeStringValue((*req)[0]->toString(false) + "STORE"));
    req->withString(other.getName());
    for (std::size_t i = req->size() - 1; i > 1; --i) {
        req->swap(i-1, *req, i);
    }
    client.callVoid(*req);
}

/*************************** DbStringSetOperation ***************************/

/** Constructor.
    \param op Operation
    \param a,b Arguments */
DbStringSetOperation::DbStringSetOperation(const char* op, DbStringSet a, DbStringSet b)
    : client(a.getClient()),
      req(new DbRequest())
{
    req->withString(op).withString(a.getName()).withString(b.getName());
}

/** Add an additional argument. */
DbStringSetOperation&
DbStringSetOperation::andAlso(DbStringSet x)
{
    req->withString(x.getName());
    return *this;
}

/** Get the result. */
void
DbStringSetOperation::get(StringList_t& list)
{
    std::auto_ptr<IntValue> result(client.call(*req));
    toStringListResult(result.get(), list);
}

/** Store the result in another set. */
void
DbStringSetOperation::storeTo(DbStringSet other)
{
    req->setNew(0, makeStringValue((*req)[0]->toString(false) + "STORE"));
    req->withString(other.getName());
    for (std::size_t i = req->size() - 1; i > 1; --i) {
        req->swap(i-1, *req, i);
    }
    client.callVoid(*req);
}
