/**
  *  \file planetscentral/talk/spam.cc
  *  \brief Spam Detection
  */

#include <stdio.h>
#include "planetscentral/talk/spam.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/smiley.h"
#include "planetscentral/osutil.h"
#include "planetscentral/talk/talk.h"

static size_t
countLinks(const string_t& text)
{
    string_t::size_type count = 0;
    string_t::size_type pos = 0;
    InlineInfo info;
    while (findInlineElement(text, pos, InlineItemSet(InlineInfo::Link), info)) {
        if (info.length == 0 || info.start < pos) {
            // cannot happen; avoid deadlock if it does anyway
            break;
        }
        pos = info.start + info.length;
        ++count;
    }
    return count;
}

/** Check for spam.
    \param subject post subject
    \param text post text
    \param time time of post
    \param u posting user
    \return true if this posting is spam */
bool
checkSpam(const string_t& /*subject*/, const string_t& text, int32_t time, User& u)
{
    // User check
    // SPAM_USER_PROFILE can be
    // - not present (=default behaviour)
    // - 0=user is excempt from spam check but does not see spam
    // - 1=user is a spammer
    // - 2=user is excempt from spam check and can see spam
    DbHash profile = u.profile();
    if (profile.hasField(SPAM_USER_PROFILE)) {
        if (profile.intField(SPAM_USER_PROFILE).get() == 1) {
            // recognized spammer
            MutexGuard g(output_mutex);
            printf("[spam] marking posting as spam because user is a spammer\n");
            return true;
        } else {
            // user is excempt from spam check
            return false;
        }
    }

    // Compute score
    int score = 0;

    // - Chinese language
    // Matches 22/22 spams, 8/8 spammers, with no false positives.
    // Question is how racist we want to be.
    if (strUCase(profile.getStringField("createacceptlanguage")).substr(0, 2) == "ZH") {
        score += 20;
    }

    // - Freemail provider.
    // Matches 22/22 spams, 8/8 spammers, but @hotmail has some false positives.
    string_t mail(strLCase(profile.getStringField("email")));
    if (mail.find("@outlook") != string_t::npos || mail.find("@hotmail") != string_t::npos) {
        score += 10;
    }

    // - Post age
    // 8/8 spammers send their first spam within 15 minutes, 6/8 within 5 minutes.
    int32_t age = time - profile.getIntField("createtime") / 60;
    if (age <= 5) {
        score += 5;
    }
    if (age <= 15) {
        score += 5;
    }
    if (age <= 60) {
        score += 5;
    }

    // - Post size
    // 22/22 spams are over 3k, which is pretty large for a user's initial posting.
    // Use 2.5k for some margin.
    if (text.size() > 2500) {
        score += 10;
    }

    // - Links
    // 20/22 spams have 3 or more links.
    if (countLinks(text) >= 3) {
        score += 20;
    }

    // Recorded spams yield values between 55 and 75 here.
    // Set the margin at 65 for now.
    if (score >= 65) {
        // This is spam.
        MutexGuard g(output_mutex);
        printf("[spam] marking post+user as spam, score is %d\n", score);
        profile.setField(SPAM_USER_PROFILE, 1);
        return true;
    } else if (score >= 20) {
        // Log weak candidates, just for analysis
        MutexGuard g(output_mutex);
        printf("[spam] score is %d\n", score);
        return false;
    } else {
        // No indication of spam.
        return false;
    }
}
