/**
  *  \file planetscentral/talk/syntaxc.cc
  *  \brief Talk: Syntax Highlighter for C/C++/Java/JavaScript files
  */

#include <cstddef>
#include "planetscentral/talk/syntaxc.h"
#include "cpluslib/misc.h"

namespace {
    enum Class {
        cWhitespace = 1,        // horizontal whitespace
        cNewline    = 2,        // newline
        cSlash      = 4,        // slash
        cStar       = 8,        // *
        cHash       = 16,       // #
        cQuote      = 32,       // " or '
        cLetter     = 64,       // alpha or _
        cDigit      = 128,      // 0-9
        cPunct      = 256,      // punctuation that can appear before a regexp: binary operators, "("
        cOther      = 512,      // punctuation
    };

    Class classify(char c) {
        switch (c) {
         case ' ':
         case '\t':
         case '\v':
         case '\f':
            return cWhitespace;

         case '\r':
         case '\n':
            return cNewline;

         case '/':
            return cSlash;

         case '*':
            return cStar;

         case '#':
            return cHash;

         case '"':
         case '\'':
            return cQuote;

         case '[':
         case '(':
         case ',':
         case '=':
         case ':':
         case '!':
         case '&':
         case '|':
         case '?':
         case '{':
         case '}':
         case ';':
         case '-':
         case '+':
         case '<':
         case '>':
         case '~':
            return cPunct;

         default:
            if (c >= '0' && c <= '9') {
                return cDigit;
            } else if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || c == '_') {
                return cLetter;
            } else {
                return cOther;
            }
        }
    }

    void skipContinuation(const string_t& str, string_t::size_type& n)
    {
        while (str.size() - n >= 2
               && str[n] == '\\'
               && (str[n+1] == '\r' || str[n+1] == '\n'))
        {
            if (str[n+1] == '\r' && str.size() - n >= 3 && str[n+2] == '\n') {
                ++n;
            }
            n += 2;
        }
    }

    string_t::size_type skip(const string_t& str, string_t::size_type& pos, int c)
    {
        string_t::size_type n = pos;
        while (n < str.size() && (classify(str[n]) & c) != 0) {
            ++n;
            skipContinuation(str, n);
        }

        string_t::size_type result = n - pos;
        pos = n;
        return result;
    }

    bool skip1(const string_t& str, string_t::size_type& pos, int c)
    {
        bool result = false;
        if (pos < str.size() && (classify(str[pos]) & c) != 0) {
            result = true;
            ++pos;
        }
        return result;
    }

    string_t getIdentifier(const string_t& str, string_t::size_type beg, string_t::size_type end)
    {
        string_t result;
        skipContinuation(str, beg);
        while (beg < end) {
            result += str[beg++];
            skipContinuation(str, beg);
        }
        return result;
    }

    void skipRegexp(const string_t& str, string_t::size_type& pos)
    {
        bool quote = false;
        bool bracket = false;
        while (pos < str.size()) {
            char ch = str[pos++];
            if (quote) {
                quote = false;
            } else if (ch == '\\') {
                quote = true;
            } else if (ch == '[') {
                bracket = true;
            } else if (ch == ']') {
                bracket = false;
            } else if (bracket || ch != '/') {
                // part of regexp
            } else {
                break;
            }
        }
    }

    /*
     *  Keyword list
     *
     *  Keywords have been taken from
     *    C99 (iso-9899-1999.pdf)
     *    C++11 (n3242.pdf)
     *    Java5 (langspec-3.0.pdf)
     *    ES5 (ECMA-262 (ECMAScript 5th edition).pdf)
     */
    struct Keyword {
        const char* word;
        int where;
    };
    static const Keyword keywords[] = {
        { "_Bool",            SyntaxC::LangC },                                                      // C99
        { "_Complex",         SyntaxC::LangC  },                                                     // C99
        { "_Imaginary",       SyntaxC::LangC   },                                                    // C99
        { "abstract",         SyntaxC::LangJava  },                                                  // Java5
        { "alignas",          SyntaxC::LangCXX },                                                    // C++11
        { "alignof",          SyntaxC::LangCXX },                                                    // C++11
        { "and",              SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "and_eq",           SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "asm",              SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (but common on C)
        { "assert",           SyntaxC::LangC | SyntaxC::LangCXX | SyntaxC::LangJava },               // Java5
        { "auto",             SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11, C99
        { "bitand",           SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "bitor",            SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "bool",             SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "boolean",          SyntaxC::LangJava },                                                   // Java5
        { "break",            -1 },                                                                  // common
        { "byte",             SyntaxC::LangJava },                                                   // Java5
        { "case",             -1 },                                                                  // common
        { "catch",            SyntaxC::LangCXX | SyntaxC::LangJava | SyntaxC::LangJavaScript },      // C++11, ES5, Java5
        { "char",             SyntaxC::LangC | SyntaxC::LangCXX | SyntaxC::LangJava },               // C++11, C99, Java5
        { "char16_t",         SyntaxC::LangCXX  },                                                   // C++11
        { "char32_t",         SyntaxC::LangCXX  },                                                   // C++11
        { "class",            SyntaxC::LangCXX | SyntaxC::LangJava | SyntaxC::LangJavaScript },      // C++11, ES5 (future), Java5
        { "compl",            SyntaxC::LangCXX },                                                    // C++11
        { "complex",          SyntaxC::LangC },                                                      // C99 (pseudo)
        { "const",            -1 },                                                                  // common (future reserved word in ES5)
        { "const_cast",       SyntaxC::LangCXX },                                                    // C++11
        { "constexpr",        SyntaxC::LangCXX  },                                                   // C++11
        { "continue",         -1 },                                                                  // common
        { "debugger",         SyntaxC::LangJavaScript },                                             // ES5
        { "decltype",         SyntaxC::LangCXX },                                                    // C++11
        { "default",          -1 },                                                                  // common
        { "delete",           SyntaxC::LangCXX | SyntaxC::LangJavaScript },                          // C++11
        { "do",               -1 },                                                                  // common
        { "double",           SyntaxC::LangCXX | SyntaxC::LangC | SyntaxC::LangJava },               // C++11, C99, Java5
        { "dynamic_cast",     SyntaxC::LangCXX },                                                    // C++11
        { "else",             -1 },                                                                  // common
        { "enum",             -1 },                                                                  // common (future reserved word in ES5)
        { "explicit",         SyntaxC::LangCXX },                                                    // C++11
        { "export",           SyntaxC::LangCXX | SyntaxC::LangJavaScript },                          // C++11, ES5 (future)
        { "extends",          SyntaxC::LangJavaScript | SyntaxC::LangJava },                         // ES5 (future), Java5
        { "extern",           SyntaxC::LangCXX | SyntaxC::LangC },                                   // C++11, C99
        { "false",            -1 },                                                                  // common (macro in C)
        { "final",            SyntaxC::LangJava },                                                   // Java5
        { "finally",          SyntaxC::LangJava | SyntaxC::LangJavaScript },                         // ES5, Java5
        { "float",            SyntaxC::LangCXX | SyntaxC::LangC | SyntaxC::LangJava },               // C++11, C99, Java5
        { "for",              -1 },                                                                  // common
        { "friend",           SyntaxC::LangCXX },                                                    // C++11
        { "function",         SyntaxC::LangJavaScript },                                             // ES5
        { "goto",             SyntaxC::LangC | SyntaxC::LangCXX | SyntaxC::LangJava },               // C++11, C99, Java5
        { "if",               -1 },                                                                  // common
        { "implements",       SyntaxC::LangJava | SyntaxC::LangJavaScript  },                        // ES5 (future strict), Java5
        { "import",           SyntaxC::LangJava | SyntaxC::LangJavaScript },                         // ES5 (future), Java5
        { "in",               SyntaxC::LangJavaScript },                                             // ES5
        { "inline",           SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11, C99
        { "instanceof",       SyntaxC::LangJavaScript | SyntaxC::LangJava },                         // ES5, Java5
        { "int",              SyntaxC::LangC | SyntaxC::LangCXX | SyntaxC::LangJava },               // C++11, C99, Java5
        { "interface",        SyntaxC::LangJavaScript | SyntaxC::LangJava },                         // ES5 (future strict), Java5
        { "let",              SyntaxC::LangJavaScript },                                             // ES5 (future strict)
        { "long",             SyntaxC::LangCXX | SyntaxC::LangC | SyntaxC::LangJava },               // C++11, C99, Java5
        { "mutable",          SyntaxC::LangCXX },                                                    // C++11
        { "namespace",        SyntaxC::LangCXX },                                                    // C++11
        { "native",           SyntaxC::LangJava },                                                   // Java5
        { "new",              SyntaxC::LangCXX | SyntaxC::LangJavaScript | SyntaxC::LangJava },      // C++11, ES5, Java5
        { "noexcept",         SyntaxC::LangCXX },                                                    // C++11
        { "not",              SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "not_eq",           SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "null",             SyntaxC::LangJava | SyntaxC::LangJavaScript },                         // ES5
        { "nullptr",          SyntaxC::LangCXX },                                                    // C++11
        { "operator",         SyntaxC::LangCXX },                                                    // C++11
        { "or",               SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "or_eq",            SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "package",          SyntaxC::LangJava | SyntaxC::LangJavaScript },                         // ES5 (future strict), Java5
        { "private",          SyntaxC::LangCXX | SyntaxC::LangJavaScript | SyntaxC::LangJava },      // C++11, ES5 (future strict), Java5
        { "protected",        SyntaxC::LangCXX | SyntaxC::LangJavaScript | SyntaxC::LangJava },      // C++11, ES5 (future strict), Java5
        { "public",           SyntaxC::LangCXX | SyntaxC::LangJavaScript | SyntaxC::LangJava },      // C++11, ES5 (future strict), Java5
        { "register",         SyntaxC::LangCXX | SyntaxC::LangC },                                   // C++11,C99
        { "reinterpret_cast", SyntaxC::LangCXX },                                                    // C++11
        { "restrict",         SyntaxC::LangC },                                                      // C99
        { "return",           -1 },                                                                  // common
        { "short",            SyntaxC::LangCXX | SyntaxC::LangC | SyntaxC::LangJava },               // C++11, C99, Java5
        { "signed",           SyntaxC::LangCXX | SyntaxC::LangC },                                   // C++11, C99
        { "sizeof",           SyntaxC::LangCXX | SyntaxC::LangC },                                   // C++11, C99
        { "static",           -1 },                                                                  // common (future strict reserved word in ES5)
        { "static_assert",    SyntaxC::LangCXX },                                                    // C++11
        { "static_cast",      SyntaxC::LangCXX },                                                    // C++11
        { "strictfp",         SyntaxC::LangJava },                                                   // Java5
        { "struct",           SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11, C99
        { "super",            SyntaxC::LangJava | SyntaxC::LangJavaScript },                         // ES5 (future), Java5
        { "switch",           -1 },                                                                  // common
        { "synchronized",     SyntaxC::LangJava },                                                   // Java5
        { "template",         SyntaxC::LangCXX },                                                    // C++11
        { "this",             SyntaxC::LangCXX | SyntaxC::LangJava | SyntaxC::LangJavaScript },      // C++11, ES5, Java5
        { "thread_local",     SyntaxC::LangCXX },                                                    // C++11
        { "throw",            SyntaxC::LangCXX | SyntaxC::LangJava | SyntaxC::LangJavaScript },      // C++11, ES5, Java5
        { "throws",           SyntaxC::LangJava },                                                   // Java5
        { "transient",        SyntaxC::LangJava },                                                   // Java5
        { "true",             -1 },                                                                  // common (macro in C)
        { "try",              SyntaxC::LangCXX | SyntaxC::LangJavaScript | SyntaxC::LangJava },      // C++11, ES5, Java5
        { "typedef",          SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11, C99
        { "typeid",           SyntaxC::LangCXX },                                                    // C++11
        { "typename",         SyntaxC::LangCXX },                                                    // C++11
        { "typeof",           SyntaxC::LangJavaScript },                                             // ES5
        { "union",            SyntaxC::LangCXX | SyntaxC::LangC },                                   // C++11, C99
        { "unsigned",         SyntaxC::LangCXX | SyntaxC::LangC },                                   // C++11
        { "using",            SyntaxC::LangCXX },                                                    // C++11
        { "var",              SyntaxC::LangJavaScript },                                             // ES5
        { "virtual",          SyntaxC::LangCXX },                                                    // C++11
        { "void",             -1 },                                                                  // common
        { "volatile",         SyntaxC::LangCXX | SyntaxC::LangC | SyntaxC::LangJava },               // C++11, C99, Java5
        { "wchar_t",          SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (typedef in C)
        { "while",            -1 },                                                                  // common
        { "with",             SyntaxC::LangJavaScript },                                             // ES5
        { "xor",              SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "xor_eq",           SyntaxC::LangC | SyntaxC::LangCXX },                                   // C++11 (macro in C)
        { "yield",            SyntaxC::LangJavaScript },                                             // ES5 (future strict)
    };

    int findKeyword(const string_t str)
    {
        std::size_t idx = 0;
        for (std::size_t i = 0x80; i > 0; i >>= 1) {
            std::size_t nidx = idx | i;
            if (nidx < countof(keywords) && str >= keywords[nidx].word) {
                idx = nidx;
            }
        }
        if (keywords[idx].word == str) {
            return keywords[idx].where;
        } else {
            return 0;
        }
    }
}

SyntaxC::SyntaxC(int language)
    : language(language),
      str(),
      pos(0),
      state(sDefault),
      accept_regexp(false),
      accept_preproc(true)
{ }

SyntaxC::~SyntaxC()
{ }

void
SyntaxC::init(const string_t& str)
{
    this->str            = str;
    this->pos            = 0;
    this->state          = sDefault;
    this->accept_regexp  = false;
    this->accept_preproc = true;
}

bool
SyntaxC::scan(Segment& result)
{
    // Whitespace always accepted
    result.start(pos);
    if (skip(str, pos, cWhitespace)) {
        result.finish(DefaultFormat, pos);
        return true;
    }

    // Special states
    if (state == sMacroName) {
        if (skip(str, pos, cLetter)) {
            skip(str, pos, cLetter | cDigit);
            result.finish(NameFormat, pos);
            state          = sDefault;
            accept_regexp  = false;
            accept_preproc = false;
            return true;
        } else {
            state = sDefault;
            accept_regexp  = false;
            accept_preproc = false;
        }
    }

    if (state == sIncludeFileName) {
        if (pos < str.size() && str[pos] == '<') {
            while (pos < str.size() && classify(str[pos]) != cNewline && str[pos] != '>') {
                ++pos;
                skipContinuation(str, pos);
            }
            if (pos < str.size() && str[pos] == '>') {
                ++pos;
                skipContinuation(str, pos);
            }
            result.finish(StringFormat, pos);
            state          = sDefault;
            accept_regexp  = false;
            accept_preproc = false;
            return true;
        } else {
            state = sDefault;
            accept_regexp  = false;
            accept_preproc = false;
        }
    }

    // Default state
    if (skip(str, pos, cNewline)) {
        // Newline
        result.finish(DefaultFormat, pos);
        accept_preproc = true;
        return true;
    } else if (skip1(str, pos, cSlash)) {
        // Slash. Could be comment or regexp.
        skipContinuation(str, pos);
        if (skip1(str, pos, cSlash)) {
            // "//" comment
            skipContinuation(str, pos);
            skip(str, pos, ~cNewline);
            result.finish(CommentFormat, pos);
            return true;
        } else if (skip1(str, pos, cStar)) {
            // "/*" comment
            skipContinuation(str, pos);
            while (1) {
                skip(str, pos, ~cStar);
                if (!skip(str, pos, cStar)) {
                    // End of file reached
                    break;
                }
                if (skip1(str, pos, cSlash)) {
                    // End of comment
                    skipContinuation(str, pos);
                    break;
                }
            }
            result.finish(CommentFormat, pos);
            return true;
        } else if (accept_regexp && (language & LangJavaScript) != 0) {
            // Regexp
            skipRegexp(str, pos);
            result.finish(StringFormat, pos);
            accept_regexp  = false;
            accept_preproc = false;
            return true;
        } else {
            // Just a slash
            result.finish(DefaultFormat, pos);
            accept_regexp = true;
            accept_preproc = false;
            return true;
        }
    } else if (accept_preproc && (language & (LangC | LangCXX)) != 0 && skip1(str, pos, cHash)) {
        // Preprocessor
        skipContinuation(str, pos);
        skip(str, pos, cWhitespace);
        string_t::size_type mark = pos;
        skip(str, pos, cLetter | cDigit);
        string_t ident = getIdentifier(str, mark, pos);
        if (ident == "ifdef" || ident == "undef" || ident == "define" || ident == "ifndef") {
            state = sMacroName;
        } else if (ident == "include" || ident == "import") {
            state = sIncludeFileName;
        } else {
            // keep state
        }
        result.finish(SectionFormat, pos);
        accept_regexp = false;
        accept_preproc = false;
        return true;
    } else if (skip1(str, pos, cQuote)) {
        // String
        char delim = str[pos-1];
        bool quoted = false;
        skipContinuation(str, pos);
        while (pos < str.size()) {
            char ch = str[pos++];
            skipContinuation(str, pos);
            if (quoted) {
                quoted = false;
            } else if (ch == '\\') {
                quoted = true;
            } else if (ch != delim) {
                quoted = false;
            } else {
                break;
            }
        }
        result.finish(StringFormat, pos);
        accept_regexp  = false;
        accept_preproc = false;
        return true;
    } else if (skip(str, pos, cDigit)) {
        // Number
        result.finish(DefaultFormat, pos);
        accept_regexp  = false;
        accept_preproc = false;
        return true;
    } else if (skip(str, pos, cLetter)) {
        // Identifier
        skip(str, pos, cLetter | cDigit);
        if ((findKeyword(getIdentifier(str, result.start_pos, pos)) & language) != 0) {
            result.finish(KeywordFormat, pos);
        } else {
            result.finish(DefaultFormat, pos);
        }
        accept_regexp  = false;
        accept_preproc = false;
        return true;
    } else if (skip(str, pos, cPunct | cStar)) {
        // Punctuation
        result.finish(DefaultFormat, pos);
        accept_regexp  = true;
        accept_preproc = false;
        return true;
    } else if (skip(str, pos, cOther | cHash)) {
        // More punctuation
        result.finish(DefaultFormat, pos);
        accept_regexp  = false;
        accept_preproc = false;
        return true;
    } else {
        return false;
    }
}
