/**
  *  \file planetscentral/talk/syntaxini.cc
  *  \brief Talk: Syntax Highlighter for INI files
  */

#include <cctype>
#include "planetscentral/talk/syntaxini.h"
#include "cpluslib/string.h"
#include "planetscentral/talk/syntaxdb.h"

namespace {
    enum Class {
        cSpace    = 1,
        cComment  = 2,
        cLBracket = 4,
        cRBracket = 8,
        cPercent  = 16,
        cEqual    = 32,
        cNewline  = 64,
        cOther    = 128
    };

    Class classify(char c) {
        switch (c) {
         case ' ':
         case '\t':
         case '\v':
         case '\f':
            return cSpace;

         case '\r':
         case '\n':
            return cNewline;

         case '#':
         case ';':
            return cComment;

         case '[':
            return cLBracket;

         case ']':
            return cRBracket;

         case '%':
            return cPercent;

         case '=':
            return cEqual;

         default:
            return cOther;
        }
    }

    string_t::size_type skip(const string_t& str, string_t::size_type& pos, int c)
    {
        string_t::size_type n = pos;
        while (n < str.size() && (classify(str[n]) & c) != 0) {
            ++n;
        }

        string_t::size_type result = n - pos;
        pos = n;
        return result;
    }

    bool skip1(const string_t& str, string_t::size_type& pos, int c)
    {
        bool result = false;
        if (pos < str.size() && (classify(str[pos]) & c) != 0) {
            result = true;
            ++pos;
        }
        return result;
    }

}


/******************************* SyntaxIni *******************************/

/** Constructor.
    \param db Syntax database
    \param defaultSection Start interpreting keys in this section */
SyntaxIni::SyntaxIni(const SyntaxDatabase& db, string_t defaultSection)
    : db(db),
      section(defaultSection),
      str(),
      pos(0),
      state(sBOL)
{ }

SyntaxIni::~SyntaxIni()
{ }

void
SyntaxIni::init(const string_t& str)
{
    this->str = str;
    this->pos = 0;
    this->state = sBOL;
}

bool
SyntaxIni::scan(Segment& result)
{
    // End reached?
    if (pos >= str.size()) {
        return false;
    }

    // Dispatch depending on state
    switch (state) {
     case sBOL:
        result.start(pos);
        if (skip(str, pos, cSpace | cNewline)) {
            // Whitespace
            result.finish(DefaultFormat, pos);
        } else if (skip1(str, pos, cComment)) {
            // Comment
            Format fmt = pos < str.size() && str[pos] == str[pos-1] ? Comment2Format : CommentFormat;
            skip(str, pos, ~cNewline);
            result.finish(fmt, pos);
        } else if (skip1(str, pos, cLBracket)) {
            // "["
            string_t::size_type a = pos;
            skip(str, pos, ~(cNewline | cRBracket));
            section = strTrim(string_t(str, a, pos-a));
            if (skip1(str, pos, cRBracket)) {
                state = sAfterSection;
            }
            result.finish(SectionFormat, pos);
        } else if (skip1(str, pos, cPercent)) {
            // "%"
            skip(str, pos, cSpace);
            string_t::size_type a = pos;
            skip(str, pos, ~(cNewline | cSpace));
            section = strTrim(string_t(str, a, pos-a));
            result.finish(SectionFormat, pos);
            if (!skip1(str, pos, cNewline)) {
                state = sAfterSection;
            }
        } else if (string_t::size_type n = skip(str, pos, cOther | cLBracket | cRBracket)) {
            // Name. Accept [] as well for things like "foo[1] = ...", maybe someone uses that.
            string_t key(strLTrim(string_t(str, pos-n, n)));
            if (skip1(str, pos, cNewline)) {
                result.finish(DefaultFormat, pos);
            } else {
                result.finish(NameFormat, pos);
                state = sAfterName;

                // Links
                string_t pfx = "ini.";
                string_t::size_type dot = key.find('.');
                if (dot != string_t::npos
                    && (section.empty()
                        || (dot == section.size() && key.compare(0, dot, section) == 0)))
                {
                    pfx += key;
                } else{ 
                    pfx += section;
                    pfx += ".";
                    pfx += key;
                }
                result.link = db.get(pfx + ".link");
                result.info = db.get(pfx + ".info");
            }
        } else {
            // Don't know what it is. Skip the whole line.
            skip(str, pos, ~cNewline);
            skip1(str, pos, cNewline);
            result.finish(DefaultFormat, pos);
        }
        break;

     case sAfterSection:
        result.start(pos);
        if (skip(str, pos, cSpace)) {
            // At least some spaces
            if (skip1(str, pos, cNewline)) {
                state = sBOL;
            }
            result.finish(DefaultFormat, pos);
        } else if (skip(str, pos, cComment)) {
            // A comment
            skip(str, pos, ~cNewline);
            if (skip1(str, pos, cNewline)) {
                state = sBOL;
            }
            result.finish(CommentFormat, pos);
        } else {
            // Something else
            skip(str, pos, ~cNewline);
            if (skip1(str, pos, cNewline)) {
                state = sBOL;
            }
            result.finish(DefaultFormat, pos);
        }
        break;
        
     case sAfterName:
        result.start(pos);
        skip(str, pos, ~cNewline);
        if (skip1(str, pos, cNewline)) {
            state = sBOL;
        }
        result.finish(DefaultFormat, pos);
        break;
    }
    return true;
}

#ifdef TEST_HLINI
#include <iostream>
int main()
{
    // Read
    string_t tmp, total;
    while (getline(std::cin, tmp)) {
        total += tmp;
        total += '\n';
    }

    // Parse
    SyntaxIni hl("");
    hl.init(total);

    SyntaxHighlighter::Segment seg;
    while (hl.scan(seg)) {
        string_t segstr(total, seg.start_pos, seg.length);
        std::printf("%3d %5d %5d: \"%s\"\n", seg.fmt, seg.start_pos, seg.length, segstr.c_str());
    }
}
#include "planetscentral/talk/syntax.cc"
#endif
